from . import Resolver, RegistartionFilter
from .registration_filters import all_registrations


def use_registered(cls: type, filter: RegistartionFilter = all_registrations):
    def factory(resolver: Resolver):
        return resolver.resolve(cls, filter=filter)

    return factory



from typing import _GenericAlias  # type: ignore
from typing import Generic, TypeVar
from collections.abc import Callable
from typing import get_type_hints
from queue import Queue
import inspect


def get_subclasses(cls: type, filter: Callable[[type], bool] = lambda t: True):
    queue = Queue()
    queue.put(cls)
    items = []
    while not queue.empty():
        t = queue.get_nowait()

        for sub in t.__subclasses__():
            queue.put(sub)

            if filter(sub):
                items.append(sub)

    return items


def get_generic_bases(cls: type, filter: Callable[[type], bool] = lambda t: True):
    queue = Queue()
    queue.put(cls)
    items = []
    while not queue.empty():
        t = queue.get_nowait()

        for sub in getattr(t, "__orig_bases__", ()):
            queue.put(sub)
            if orig := getattr(sub, "__origin__", None):
                queue.put(orig)

            if filter(sub):
                items.append(sub)

    return items


def is_generic_type_closed(cls: type):
    m = get_typevar_to_type_mapping(cls)
    for k, v in m.items():
        if k == v:
            return False
    return True


def get_type_args(cls: type):
    return getattr(cls, "__args__", ())


def is_open_generic_type(cls: type):
    type_args = get_type_args(cls)
    return any([isinstance(t, TypeVar) for t in type_args])


def get_generic_type_args(type: type):
    queue = Queue()
    queue.put(type)

    while not queue.empty():
        type_check = queue.get()
        if getattr(type_check, "__origin__", None) == Generic:
            return type_check.__args__

        for base in getattr(type_check, "__orig_bases__", ()):
            queue.put(base)
            queue.put(base.__origin__)

    return ()


## OLD
def get_closed_open_generic_alias(cls: type):
    queue = Queue()
    queue.put(cls)

    while not queue.empty():
        type_check = queue.get()
        if type(type_check) == _GenericAlias:
            return type_check

        for base in getattr(type_check, "__orig_bases__", ()):
            queue.put(base)
            queue.put(base.__origin__)

    return None


def get_first_open_generic_alias(cls: type):
    queue = Queue()
    queue.put(cls)

    if root_origin := (getattr(cls, "__origin__", None)):
        queue.put(root_origin)

    while not queue.empty():
        type_check = queue.get()
        if getattr(type_check, "__origin__", None) == Generic:
            return type_check

        for base in getattr(type_check, "__orig_bases__", ()):
            queue.put(base)
            if orig := getattr(base, "__origin__", None):
                queue.put(orig)

    return None


## NEW
def get_open_generic_aliases(cls: type):
    queue = Queue()
    queue.put(cls)
    aliases = []

    if root_origin := (getattr(cls, "__origin__", None)):
        queue.put(root_origin)

    while not queue.empty():
        type_check = queue.get()
        if getattr(type_check, "__origin__", None) == Generic:
            aliases.append(type_check)

        for base in getattr(type_check, "__orig_bases__", ()):
            queue.put(base)
            if orig := getattr(base, "__origin__", None):
                queue.put(orig)

    return aliases


def get_generic_aliases(cls: type):
    queue = Queue()
    queue.put(cls)
    if orig := getattr(cls, "__origin__", None):
        queue.put(orig)

    aliases = []
    while not queue.empty():
        type_check = queue.get()
        if type(type_check) == _GenericAlias:
            aliases.append(type_check)

        for base in getattr(type_check, "__orig_bases__", ()):
            queue.put(base)
            if base_orig := getattr(base, "__origin__", None):
                queue.put(base_orig)

    return aliases


def get_typevar_to_type_mapping(cls: type) -> dict[type, type]:
    generic_definitions = get_open_generic_aliases(cls)
    generic_implementations = get_generic_aliases(cls)

    generic_implementations = generic_implementations[: len(generic_definitions)]

    generic_definitions.reverse()
    generic_implementations.reverse()

    mapping = {}

    for idx, definition in enumerate(generic_definitions):
        implementation = generic_implementations[idx]
        additions = dict(zip(definition.__args__, implementation.__args__))
        mapping = {**mapping, **additions}

    return mapping


def get_generic_types(cls: type):
    mapping = get_typevar_to_type_mapping(cls)
    return tuple(mapping.values())


def try_to_complete_generic(open_type: _GenericAlias, closed_type: type) -> type:
    if is_generic_type_closed(open_type):
        return open_type

    mapping = get_typevar_to_type_mapping(closed_type)
    new_args = tuple([mapping.get(a, a) for a in open_type.__args__])
    return open_type[new_args]


def get_type_args_for_function(fn: Callable) -> dict[str, type]:
    return {k: v for k, v in get_type_hints(fn).items() if k != "return"}


def get_generic_type(obj):
    """Get the generic type of an object if possible, or runtime class otherwise.
    Examples::
        class Node(Generic[T]):
            ...
        type(Node[int]()) == Node
        get_generic_type(Node[int]()) == Node[int]
        get_generic_type(Node[T]()) == Node[T]
        get_generic_type(1) == int
    """

    gen_type = getattr(obj, "__orig_class__", None)
    return gen_type if gen_type is not None else type(obj)



from . import Registration
from .functional_utils import constant

all_registrations = constant(True)


def with_name(name: str):
    """
    Filter registartions equal the name
    """

    def predicate(r: Registration):
        return r.name == name

    return predicate


def name_starts_with(prefix: str):
    """
    Filter registartions where the name starts the prefix
    """

    def predicate(r: Registration):
        if r.name is not None:
            return r.name.startswith(prefix)
        return False

    return predicate


def name_ends_with(suffix: str):
    """
    Filter registartions where the name ends the suffix
    """

    def predicate(r: Registration):
        if r.name is not None:
            return r.name.endswith(suffix)
        return False

    return predicate


def is_named(r: Registration):
    """
    Filter registartions that have a name
    """
    return r.is_named


def with_implementation(implementation: type):
    """
    Filter to registartions that have the implementation
    """

    def predicate(r: Registration):
        return r.implementation == implementation

    return predicate


def has_generic_args_matching(pair: tuple[type, type]):
    """
    Filter registartions where generic type arg matches
    >>> TBar = TypeVar("TBar")
    >>> class Foo(Generic[TBar]):
    ...    pass
    >>> class IntFoo(Foo[int]):
    ...    pass
    >>> class StringFoo(Foo[str]):
    ...    pass
    >>> container.register(Foo, IntFoo)
    >>> container.register(Foo, StringFoo)
    >>> container.resolve(Foo, filter=has_generic_args_matching((TBar, int))) # returns instance of IntFoo
    """

    def predicate(r: Registration):
        return r.generic_mapping.get(pair[0]) == pair[1]

    return predicate



"""Simple IOC container.
"""
from __future__ import annotations
import abc
from collections import defaultdict, deque
from dataclasses import dataclass
import sys
import types
from enum import IntEnum
from typing import Any, ForwardRef, Type, get_type_hints
from collections.abc import Callable
from typing import _GenericAlias  # type: ignore
from uuid import uuid4
from .functional_utils import constant, fn_and, fn_not
from .type_filters import is_abstract
from .typing_utils import (
    get_generic_bases,
    get_generic_types,
    get_subclasses,
    get_typevar_to_type_mapping,
    try_to_complete_generic,
)
import inspect


class _empty:
    def __bool__(self):
        return False


class ArgInfo:
    def __init__(self, name: str, arg_type: type, default_value: Any):
        self.name = name
        self.arg_type = arg_type
        self.default_value = (
            _empty if default_value == inspect._empty else default_value
        )


def get_arg_info(subject: Callable) -> dict[str, ArgInfo]:
    arg_spec_fn = subject if inspect.isfunction(subject) else subject.__init__
    args = get_type_hints(arg_spec_fn, None, {})
    signature = inspect.signature(subject)
    d: dict[str, ArgInfo] = {}
    for name, param in signature.parameters.items():
        d[name] = ArgInfo(name=name, arg_type=args[name], default_value=param.default)
    return d


_all_registartions = constant(True)


class LifestyleType(IntEnum):
    transient = 0
    once_per_graph = 1
    scoped = 2
    singleton = 3


class CannotResolveException(Exception):
    def __init__(self):
        self.stack = deque()

    def append(self, d: Dependency):
        self.stack.appendleft(d)

    @staticmethod
    def print_dependency(d: Dependency):
        return f"implementation={d.parent_implementation}, dependant={{name={d.name}, type={d.service_type}}}))"

    @property
    def message(self):
        chain = ""
        horizontal_line = "\u2514\u2500\u2500>"
        vertical_line = "\u2502"
        spaces = " "

        root, *rest = self.stack

        chain += f"\n{CannotResolveException.print_dependency(root)}\n"

        for item in rest:
            chain += f"{spaces}{vertical_line}\n{spaces}{horizontal_line}{CannotResolveException.print_dependency(item)}\n"
            spaces += "     "

        return chain

    def __str__(self):
        return self.message


class Dependency:
    def __init__(
        self,
        name: str,
        parent_implementation: Callable | type,
        service_type: Any,
        settings: DependencySettings,
        default_value: Any,
    ):
        self.name = name
        self.parent_implementation = parent_implementation
        if isinstance(parent_implementation, type):
            self.service_type = try_to_complete_generic(
                service_type, parent_implementation
            )
        else:
            self.service_type = service_type
        self.settings = settings
        self.is_generic_list = getattr(self.service_type, "__origin__", None) == list
        self.default_value = default_value

    def resolve(self, context: ResolvingContext):
        if not self.settings.value == _empty:
            return self.settings.value

        if not self.default_value == _empty and self.settings.use_default_paramater:
            return self.default_value

        if self.is_generic_list:
            regs = context.find_registrations(self.service_type.__args__[0], self.settings.filter)  # type: ignore
            return [s.build(context) for s in regs]
        else:
            try:
                reg = context.find_registration(self.service_type, self.settings.filter)
                return reg.build(context)
            except CannotResolveException as ex:
                print(self.name)
                ex.append(self)
                raise ex


class RootDependency(Dependency):
    @staticmethod
    def __PARENT_ROOT__():
        pass

    def __init__(self, service_type: Any, settings: DependencySettings):
        super().__init__(
            name="__ROOT__",
            parent_implementation=RootDependency.__PARENT_ROOT__,
            service_type=service_type,
            settings=settings,
            default_value=_empty,
        )


class ImplementationCreator:
    def __init__(
        self,
        creator_function: Callable,
        dependency_config: dict[str, DependencySettings] = {},
    ):
        self.dependency_config = defaultdict(DependencySettings, dependency_config)
        self.creator_function = creator_function
        self.dependencies: dict[str, Dependency] = self._get_dependencies(
            self.creator_function, self.dependency_config
        )

    @classmethod
    def _get_default_value(cls, paramater: inspect.Parameter):
        default_value = _empty
        if not paramater.default == inspect._empty:
            default_value = paramater.default
        return default_value

    @classmethod
    def _get_dependencies(
        cls,
        creator_function: Callable,
        dependency_config: dict[str, DependencySettings],
    ) -> dict[str, Dependency]:
        # signature = inspect.signature(creator_function)
        args_infos = get_arg_info(creator_function)
        return {
            name: Dependency(
                name=name,
                parent_implementation=creator_function,
                service_type=arg_info.arg_type,
                settings=dependency_config[name],
                default_value=arg_info.default_value,
            )
            for name, arg_info in args_infos.items()
        }

    def create(self, context: ResolvingContext, **kwargs):
        for arg_name, arg_dep in self.dependencies.items():
            kwargs[arg_name] = arg_dep.resolve(context)
        built_instance = self.creator_function(**kwargs)
        return built_instance


class DecoratorCreator(ImplementationCreator):
    def __init__(self, service_type: type, decorator_type: type):
        self.service_type = service_type
        super().__init__(creator_function=decorator_type)
        self.decorated_arg = next(
            name
            for name, dep in self.dependencies.items()
            if dep.service_type == service_type
        )
        self.dependencies = {
            name: dep
            for name, dep in self.dependencies.items()
            if name != self.decorated_arg
        }


class Decorator:
    def __init__(
        self, service_type: type, decorator_type: type, filter: RegistartionFilter
    ):
        self.service_type = service_type
        self.decorator_type = decorator_type
        self.creator = DecoratorCreator(
            service_type=service_type, decorator_type=decorator_type
        )
        self.registartion_filter = filter

    def decorate(self, instance: Any, context: ResolvingContext):
        kwargs = {}
        kwargs[self.creator.decorated_arg] = instance
        return self.creator.create(context, **kwargs)


class Registration:
    def __init__(
        self,
        service_type: type,
        implementation: Callable,
        lifestyle: LifestyleType,
        name: str | None = None,
        dependency_config: dict[str, DependencySettings] = {},
    ):
        self.service_type = service_type
        self.implementation = implementation
        self.creator = ImplementationCreator(
            creator_function=implementation, dependency_config=dependency_config
        )
        self.lifestyle = lifestyle
        self.name = name
        self._id = str(uuid4())
        self.generic_mapping = get_typevar_to_type_mapping(self.service_type)

    @property
    def is_named(self):
        return self.name is not None

    def build(self, context: ResolvingContext):
        cached_instance = context.get_cached(self._id)
        if not cached_instance == _empty:
            return cached_instance
        built_instance = self.creator.create(context)

        for dec in context.find_decorators_that_apply(self):
            built_instance = dec.decorate(built_instance, context)

        context.new_instance_created(self._id, built_instance, self.lifestyle)
        return built_instance


class Registry:
    def __init__(self):
        self._registrations: dict[type, deque[Registration]] = defaultdict(deque)
        self._decorators: dict[type, deque[Decorator]] = defaultdict(deque)
        self._singletons: dict[str, Any] = {}

    def add_registration(self, registartion: Registration):
        self._registrations[registartion.service_type].appendleft(registartion)

    def add_decorator(self, decorator: Decorator):
        self._decorators[decorator.service_type].appendleft(decorator)

    def add_singleton_instance(self, registartion_id: str, instance: Any):
        self._singletons[registartion_id] = instance

    def get_registartions(self, service_type: type):
        return self._registrations[service_type]

    def get_decorators(self, service_type: type):
        return self._decorators[service_type]

    def get_singleton(self, registartion_id):
        return self._singletons.get(registartion_id)

    @property
    def singleton_items(self):
        return self._singletons.items()


class ResolvingContext:
    def __init__(self, registry: Registry, scope: Scope):
        self.registry = registry
        self.scope = scope
        self._cache = {}
        self._cache = {
            **{k: v for k, v in registry.singleton_items},
            **{k: v for k, v in scope.scoped_items},
        }

    def try_generic_fallback(self, service_type: _GenericAlias):
        return self.find_registration(service_type.__origin__, _all_registartions)

    def find_registration(
        self, service_type: type, registration_finder: Callable
    ) -> Registration:
        regs = self.find_registrations(service_type, registration_finder)
        reg = next(iter(regs), None)

        if reg is None:
            if type(service_type) == _GenericAlias:
                reg = self.try_generic_fallback(service_type)
            if reg is None:
                print(f"{service_type} is None")
                raise CannotResolveException()
        return reg

    def find_registrations(
        self, service_type: type, registration_filter: Callable[[Registration], bool]
    ) -> list[Registration]:
        scoped_registrations = [
            r
            for r in self.scope.get_registartions(service_type)
            if registration_filter(r)
        ]
        container_registrations = [
            r
            for r in self.registry.get_registartions(service_type)
            if registration_filter(r)
        ]
        return scoped_registrations + container_registrations

    def find_decorators_that_apply(self, registration: Registration):
        return [
            d
            for d in self.registry.get_decorators(registration.service_type)
            if d.registartion_filter(registration)
        ]

    def get_cached(self, reg_id: str):
        return self._cache.get(reg_id, _empty)

    def new_instance_created(
        self, reg_id: str, instance: Any, lifestyle: LifestyleType
    ):
        if lifestyle == LifestyleType.singleton:
            self.registry.add_singleton_instance(reg_id, instance)
        elif lifestyle == LifestyleType.scoped:
            self.scope.add_scoped_instance(reg_id, instance)

        if lifestyle >= LifestyleType.once_per_graph:
            self._cache[reg_id] = instance


class Resolver(abc.ABC):
    @abc.abstractmethod
    def resolve(
        self,
        cls: type,
        filter: RegistartionFilter = _all_registartions,
        *args,
        **kwargs,
    ) -> Any:
        pass


class Scope(Resolver):
    def __init__(
        self,
    ):
        self._registrations = defaultdict(deque)
        self._scoped_instances = {}

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    @abc.abstractmethod
    def add_scoped_instance(self, *args):
        pass

    @abc.abstractmethod
    def register(self, *args):
        pass

    def get_registartions(self, service_tyep):
        return self._registrations[service_tyep]

    @property
    def scoped_items(self):
        return self._scoped_instances.items()


class ContainerScope(Scope):
    def __init__(self, container: Container):
        self._container = container
        self._registrations = defaultdict(deque)
        self._scoped_instances = {}

        self.register(ContainerScope, instance=self)
        self.register(Resolver, instance=self)

    def register(
        self,
        service_type: type,
        impl_type: type | None = None,
        factory: Callable | None = None,
        instance: Any | None = None,
        name: str | None = None,
        dependency_config: dict[str, DependencySettings] = {},
    ):
        if instance is not None:
            self._registrations[service_type].appendleft(
                Registration(
                    service_type=service_type,
                    implementation=lambda: instance,
                    lifestyle=LifestyleType.scoped,
                    name=name,
                )
            )
        elif factory is not None:
            self._registrations[service_type].appendleft(
                Registration(
                    service_type=service_type,
                    implementation=factory,
                    lifestyle=LifestyleType.scoped,
                    name=name,
                    dependency_config=dependency_config,
                )
            )
        elif impl_type is not None:
            self._registrations[service_type].appendleft(
                Registration(
                    service_type=service_type,
                    implementation=impl_type,
                    lifestyle=LifestyleType.scoped,
                    name=name,
                    dependency_config=dependency_config,
                )
            )
        else:
            self._registrations[service_type].appendleft(
                Registration(
                    service_type=service_type,
                    implementation=service_type,
                    lifestyle=LifestyleType.scoped,
                    name=name,
                    dependency_config=dependency_config,
                )
            )

    def _before_start(self):
        pass

    def _after_finish(self):
        pass

    def __enter__(self):
        self._before_start()
        return self

    def __exit__(self, *args, **kwargs):
        self._after_finish()

    def get_registartions(self, service_tyep):
        return self._registrations[service_tyep]

    def add_scoped_instance(self, registration_id: str, instance: Any):
        self._scoped_instances[registration_id] = instance

    def resolve(
        self,
        service_type: type,
        filter: RegistartionFilter = _all_registartions,
    ):
        return self._container.resolve(
            service_type=service_type, filter=filter, scope=self
        )


class EmptyContainerScope(Scope):
    def __init__(self):
        super().__init__()

    def add_scoped_instance(self, *args):
        pass

    def register(self, *args):
        pass

    def resolve(self, *args):
        pass


class Container(Resolver):
    def __init__(self):
        self.registry = Registry()
        self.register(Container, instance=self)
        self.register(Resolver, instance=self)

    def register(
        self,
        service_type: type,
        impl_type: type | None = None,
        factory: Callable | None = None,
        instance: Any | None = None,
        lifestyle: LifestyleType = LifestyleType.once_per_graph,
        name: str | None = None,
        dependency_config: dict[str, DependencySettings] = {},
    ):
        if instance is not None:
            self.registry.add_registration(
                Registration(
                    service_type=service_type,
                    implementation=lambda: instance,
                    dependency_config=dependency_config,
                    lifestyle=LifestyleType.singleton,
                    name=name,
                )
            )
        elif factory is not None:
            self.registry.add_registration(
                Registration(
                    service_type=service_type,
                    implementation=factory,
                    dependency_config=dependency_config,
                    lifestyle=lifestyle,
                    name=name,
                )
            )
        elif impl_type is not None:
            self.registry.add_registration(
                Registration(
                    service_type=service_type,
                    implementation=impl_type,
                    dependency_config=dependency_config,
                    lifestyle=lifestyle,
                    name=name,
                )
            )
        else:
            self.registry.add_registration(
                Registration(
                    service_type=service_type,
                    implementation=service_type,
                    dependency_config=dependency_config,
                    lifestyle=lifestyle,
                    name=name,
                )
            )

    def register_subclasses(
        self,
        base_type: type,
        lifestyle: LifestyleType = LifestyleType.once_per_graph,
        type_filter: Callable[[type], bool] = constant(True),
    ):
        sub_classes = get_subclasses(base_type, type_filter)

        for sc in sub_classes:
            self.register(base_type, sc, lifestyle=lifestyle)
            self.register(sc, lifestyle=lifestyle)

    def register_decorator(
        self,
        service_type: type,
        decorator_type: type,
        registration_filter: Callable[[Registration], bool] = _all_registartions,
    ):
        self.registry.add_decorator(
            Decorator(
                service_type=service_type,
                decorator_type=decorator_type,
                filter=registration_filter,
            )
        )

    @staticmethod
    def _get_target_generic_base(generic_service_type: type, subclass: type):
        return next(
            (
                try_to_complete_generic(b, subclass)
                for b in get_generic_bases(
                    subclass,
                    lambda t: getattr(t, "__origin__", None) == generic_service_type,
                )
            ),
            None,
        )

    def register_open_generic(
        self,
        generic_service_type: type,
        fallback_type: type | None = None,
        lifestyle: LifestyleType = LifestyleType.once_per_graph,
        name: str | None = None,
        type_filter: Callable[[type], bool] = constant(True),
    ):
        full_type_filter = fn_and(fn_not(is_abstract), type_filter)
        subclasses = get_subclasses(generic_service_type, full_type_filter)
        for subclass in subclasses:
            target_generic_base = self._get_target_generic_base(
                generic_service_type, subclass
            )
            if target_generic_base:
                self.register(target_generic_base, subclass, lifestyle=lifestyle)

        if fallback_type:
            self.register(
                generic_service_type, fallback_type, lifestyle=lifestyle, name=name
            )

    def register_open_generic_decorator(
        self,
        generic_service_type: type,
        generic_decorator_type: type,
        type_filter: Callable[[type], bool] = constant(True),
    ):
        full_type_filter = fn_and(fn_not(is_abstract), type_filter)
        subclasses = get_subclasses(generic_service_type, full_type_filter)
        for subclass in subclasses:
            target_generic_base = self._get_target_generic_base(
                generic_service_type, subclass
            )
            if target_generic_base:
                generic_values = get_generic_types(target_generic_base)
                concrete_decorator = generic_decorator_type[generic_values]  # type: ignore
                DecoratedType = types.new_class(
                    f"DecoratedGeneric_{concrete_decorator.__name__}",
                    (concrete_decorator,),
                    {},
                )

                self.register_decorator(target_generic_base, DecoratedType)

    def resolve(
        self,
        service_type,
        filter: RegistartionFilter = _all_registartions,
        scope: Scope = EmptyContainerScope(),
    ) -> Any:
        d = RootDependency(service_type, DependencySettings(filter=filter))
        context = ResolvingContext(self.registry, scope)
        return d.resolve(context)

    def new_scope(
        self, ScopeClass: Type[ContainerScope] = ContainerScope, *args, **kwargs
    ) -> Scope:
        return ScopeClass(self, *args, **kwargs)

    def apply_module(self, module_fn: Callable[[Container], None]):
        module_fn(self)


@dataclass
class DependencySettings:
    value: Any = _empty
    use_default_paramater: bool = True
    filter: RegistartionFilter = _all_registartions


RegistartionFilter = Callable[[Registration], bool]



import inspect


def named(name: str):
    def predicate(t: type):
        return t.__name__ == name

    return predicate


def name_starts_with(name: str):
    def predicate(t: type):
        return t.__name__.startswith(name)

    return predicate


def name_end_with(name: str):
    def predicate(t: type):
        return t.__name__.endswith(name)

    return predicate


def is_in_module(*module_names: str):
    def predicate(t: type):
        return any([t.__module__ == m for m in module_names])

    return predicate


def is_abstract(t: type):
    return inspect.isabstract(t)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clean_ioc']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'clean-ioc',
    'version': '0.0.3',
    'description': 'An IOC Container for Python 3.10+',
    'long_description': '# Clean IoC\nA simple dependency injection library for python that requires nothing of your application code (except that you use typing).\n\n\n## Basic Registering and resolving\n\nThere are 4 basic modes of registering a new set of classes\n\n### Implementation\n\n```python\n\nclass UserRepository(abc.ABC)\n    @abc.abstractmethod\n    def add(self, user):\n        pass\n\nclass InMemoryUserRepository(UserRepository)\n\n    def __init__(self):\n        self.users = []\n\n    def add(self, user):\n        # This is obviously terrible, but it\'s for demo purposes\n        self.users.append(user)\n\nclass SqlAlchemyUserRepository(UserRepository)\n\n    def __init__(self):\n        # Do some db stuff here\n        pass\n\n    def add(self, user):\n        # Do some db stuff here\n        pass\n\ncontainer = Container()\ncontainer.register(UserRepository, InMemoryUserRepository)\n\n\nrepository = container.resolve(UserRepository) # This will return an InMemoryUserRepository\n\n```\n\n### Concrete Class\n\n```python\n\nclass ClientDependency\n    def get_int(self):\n        return 10\n\nclass Client\n    def __init__(self, dep: ClientDependency)\n        self.dep = dep\n\n    def get_number(self):\n        return self.dep.get_int()\n\n\ncontainer = Container()\ncontainer.register(ClientDependency)\ncontainer.register(Client)\n\nclient = container.resolve(Client)\n\nclient.get_number() # returns 10\n\n```\n\n### Factory\n\n```python\n\nclass ClientDependency\n    def get_int(self):\n        return 10\n\nclass Client\n    def __init__(self, dep: ClientDependency)\n        self.dep = dep\n\n    def get_number(self):\n        return self.dep.get_int()\n\ndef client_factory(dep: ClientDependency):\n    return Client(dep=dep)\n\n\ncontainer = Container()\ncontainer.register(ClientDependency)\ncontainer.register(Client, factory=client_factory)\n\nclient = container.resolve(Client)\n\nclient.get_number() # returns 10\n\n```\n\n\n### Instance\n\n```python\n\nclass ClientDependency\n    def __init__(self, num):\n        self.num = num\n\n    def get_int(self):\n        return self.num\n\nclass Client\n    def __init__(self, dep: ClientDependency)\n        self.dep = dep\n\n    def get_number(self):\n        return self.dep.get_int()\n\nclient_dependency = ClientDependency(num=10)\n\ncontainer = Container()\ncontainer.register(ClientDependency, instance=client_dependency)\ncontainer.register(Client)\n\nclient = container.resolve(Client)\n\nclient.get_number() # returns 10\n\n```\n\n## List resolving\n\n```python\n\nclass ClientDependency\n    def __init__(self, numbers: list[int]):\n        self.numbers = numbers\n\n    def get_numbers(self):\n        return self.numbers\n\nclass Client\n    def __init__(self, dep: ClientDependency)\n        self.dep = dep\n\n    def get_numbers(self):\n        return self.dep.get_numbers()\n\ncontainer = Container()\ncontainer.register(ClientDependency)\ncontainer.register(Client)\ncontainer.register(int, instance=1)\ncontainer.register(int, instance=2)\ncontainer.register(int, instance=3)\n\nclient = container.resolve(Client)\n\nclient.get_numbers() # returns [3, 2, 1]\n```\n\n\n## Decorators\n\nFollows a object orientated decoration pattern, rather than a decoration annotation.\nThe main reason for this was to allow decotation of registered instances\n\n```python\nclass Client\n    def __init__(self, number: int)\n        self.number = number\n\n    def get_number(self):\n        return self.number\n\n\nclass DoubleClientDecorator(Client):\n    def __init__(self, client: Client):\n        self.client = client\n    def get_number(self):\n        return self.client.get_number() * 2\n\ncontainer = Container()\n\ncontainer.register(Client)\ncontainer.register_decorator(Client, DoubleClientDecorator)\ncontainer.register(int, instance=10)\n\nclient = container.resolve(Client)\n\nclient.get_number() # returns 20\n```\n\n## Subclasses registration\n\nThis feature allows registartion of all subclasses of a giveb type\n\n```python\nclass Client(abc.ABC)\n    @abc.abstractmethod\n    def get_number(self):\n        pass\n\n\nclass TenClient(Client)\n    def get_number(self):\n        return 10\n\nclass TwentyClient(Client)\n    def get_number(self):\n        return 20\n\ncontainer = Container()\n\ncontainer.register_subclasses(Client)\n\nten_client = container.resolve(TenClient)\nten_client.get_number() # returns 10\n\ntwenty_client = container.resolve(TwentyClient)\ntwenty_client.get_number() # returns 20\n\n# Resolve all subsclasses of Client\nclient = container.resolve(list[Client]) ## [TwentyClient(), TenClient()]\n```\n\n\n## Lifestyles\nLifestyles configure how long and resolved object says alive for\nThere are 4 lifestyle types\n\n### transient\nAlways create a new instance\n\n```python\ncontainer.register(Client, lifestyle=LifestyleType.transient)\n```\n\n\n### once_per_graph (Default behaviour)\nOnly create one instance throughout the resolve call\n\n```python\ncontainer.register(Client, lifestyle=LifestyleType.once_per_graph)\n```\n\n### scoped\nOnly create a new instance through the life a scope. When not in a scope the behaviour is the same as **once_per_graph**\n\n```python\ncontainer.register(Client, lifestyle=LifestyleType.scoped)\n```\n\n### singleton\nOnly one instance of the object is created throughout the lifespan of the container\n\n```python\ncontainer.register(Client, lifestyle=LifestyleType.singleton)\n```\n\n*Note:*\nWhen registering an instance, then the behaviour is always singleton\n\n```python\ncontainer.register(int, instance=10)\n```\n\n## Open Generics\n\nRegisters all generic subclasses of the service type and allows you to resolve with the generic alias\n\n```python\nT = TypeVar("T")\n\nclass HelloCommand:\n    pass\n\nclass GoodbyeCommand:\n    pass\n\nclass CommandHandler(Generic[T]):\n    def handle(self, command: T):\n        pass\n\nclass HelloCommandHandler(CommandHandler[HelloCommand]):\n    def handle(self, command: HelloCommand):\n        print(\'HELLO\')\n\nclass GoodbyeCommandHandler(CommandHandler[GoodbyeCommand]):\n    def handle(self, command: GoodbyeCommand):\n        print(\'GOODBYE\')\n\ncontainer = Container()\ncontainer.register_open_generic(CommandHandler)\n\nh1 = container.resolve(CommandHandler[HelloCommand])\nh2 = container.resolve(CommandHandler[GoodbyeCommand])\n\nh1.handle(HelloCommand()) # prints \'HELLO\'\nh2.handle(GoodbyeCommand()) # prints \'GOODBYE\'\n\n```\n\n## Open Generic Decorators\n\n\nAllows you to add decorators to your open generic registrations\n\n```python\nT = TypeVar("T")\n\nclass HelloCommand:\n    pass\n\nclass GoodbyeCommand:\n    pass\n\nclass CommandHandler(Generic[T]):\n    def handle(self, command: T):\n        pass\n\nclass HelloCommandHandler(CommandHandler[HelloCommand]):\n    def handle(self, command: HelloCommand):\n        print(\'HELLO\')\n\nclass GoodbyeCommandHandler(CommandHandler[GoodbyeCommand]):\n    def handle(self, command: GoodbyeCommand):\n        print(\'GOODBYE\')\n\nclass AVeryBigCommandHandlerDecorator(Generic[T]):\n    def __init__(self, handler: CommandHandler[T]):\n        self.handler = handler\n\n    def handle(self, command: T):\n        print(\'A VERY BIG\')\n        self.handler.handle(command=command)\n\ncontainer = Container()\ncontainer.register_open_generic(CommandHandler)\ncontainer.register_open_generic_decorator(CommandHandler, AVeryBigCommandHandlerDecorator)\nh1 = container.resolve(CommandHandler[HelloCommand])\nh2 = container.resolve(CommandHandler[GoodbyeCommand])\n\nh1.handle(HelloCommand()) # prints \'A VERY BIG\\nHELLO\'\nh2.handle(GoodbyeCommand()) # prints \'A VERY BIG\\nGOODBYE\'\n\n```\n\n## Scopes\n\nScopes are a way to create a sub container that will live for a certain lifestyle.\nSome good use cases for scope would be for the lifespan of handling a http request with a web server or a message/event if working on a message based system\n\n\n```python\nclass Client\n    def __init__(self, number: int)\n        return number\n\n    def get_number(self):\n        return self.resolver.resolve(int)\n\ncontainer.register(int, instance=2)\n\ncontainer.register(Client)\n\nclient = container.resolve(Client)\nclient.get_number() # returns 2\n\nwith container.get_scope() as scope:\n    scope.register(int, instance=10)\n    scoped_client = scope.resolve(Client)\n    scoped_client.get_number() # returns 10\n```\n\n## Named registartions & Registartion filters\n\nBy default the last registration is what the container will return when resolve is called as below.\n\n```python\n\ncontainer = Container()\ncontainer.register(int, instance=1)\ncontainer.register(int, instance=2)\ncontainer.register(int, instance=3)\n\nnumber = container.resolve(int) # returns 3\n\n```\nTo be more selective of what we return we can add a name to the registration and apply a registartion filter when we resolve.\n\nA registartion filter is simply function that receives a **Registartion** and returns a **bool**\n\nFor example if we wanted to get the int named **"One"** we do the following\n\n```python\ncontainer = Container()\ncontainer.register(int, instance=1, name="One")\ncontainer.register(int, instance=2, name="Two")\ncontainer.register(int, instance=3, name="Three")\n\nnumber = container.resolve(int, filter=lambda r: r.name == "One") # returns 1\n```\n\nClean IOC comes with a set of in built registartion filters that can be found [here](./clean_ioc/registration_filters.py)\n\nWe can get the desired behaviour as above\n```python\nfrom clean_ioc.registartion_filters import with_name\n\ncontainer = Container()\ncontainer.register(int, instance=1, name="One")\ncontainer.register(int, instance=2, name="Two")\ncontainer.register(int, instance=3, name="Three")\n\nnumber = container.resolve(int, filter=with_name("One")) # returns 1\n```\n\n## Dependency Settings\n\nDependency settings are defined at registartion and allow you to define the selection or setting dependencies\n\n\n```python\nclass Client\n    def __init__(self, number=10)\n        self.number = number\n\n    def get_number(self):\n        return self.number\n\ncontainer = Container()\n\ncontainer.register(int, instance=1, name="One")\ncontainer.register(int, instance=2, name="Two")\n\ncontainer.register(\n    Client,\n    name="SetsValue",\n    dependency_config={"number": DependencySettings(value=50)}\n)\ncontainer.register(\n    Client,\n    name="UsesDefaultValue"\n)\ncontainer.register(\n    Client,\n    name="IgnoresDefaultParameterValue",\n    dependency_config={"number": DependencySettings(use_default_paramater=False)}\n)\ncontainer.register(\n    Client,\n    name="UsesRegistartionFilter",\n    dependency_config={"number": DependencySettings(use_default_paramater=False, filter=with_name("One"))}\n)\n\nclient1 = container.resolve(Client, filter=with_name("SetsValue"))\nclient2 = container.resolve(Client, filter=with_name("UsesDefaultValue"))\nclient3 = container.resolve(Client, filter=with_name("IgnoresDefaultParameterValue"))\nclient4 = container.resolve(Client, filter=with_name("UsesRegistartionFilter"))\n\n\nclient1.get_number() # returns 50\nclient2.get_number() # returns 10\nclient3.get_number() # returns 2\nclient4.get_number() # returns 1\n```\n\nThe order of a dependant value is as follows\n1. Setting the dependency value explicitly\n    ```python\n    DependencySettings(value=50)\n    ```\n2. Using the default parameter value if it exisis the dependency value explicitly\n    ```python\n    class Client\n    def __init__(self, number=10)\n        self.number = number\n    ```\n    If you don\'t want to use the default parameter value you can set it to false in the dependency setting\n    ```python\n    DependencySettings(use_default_paramater=False)\n    ```\n3. Going to the container registry to find a registartion using the registration filter if, if there is a default value on the dependant paramater you must explicity set.\n\n## Accessing the Container, Scope and Resolver within dependencies\n\nAccessing container directly\n\n```python\nclass Client\n    def __init__(self, container: Container)\n        self.container = container\n\n    def get_number(self):\n        return self.container.resolve(int)\n\ncontainer.register(int, instance=2)\n\ncontainer.register(Client)\n\nclient = container.resolve(Client)\nclient.get_number() # returns 2\n```\n\nAccessing Resolver also returns the container\n\n```python\n\nclass Client\n    def __init__(self, resolver: Resolver)\n        self.resolver = resolver\n\n    def get_number(self):\n        return self.resolver.resolve(int)\n\ncontainer.register(int, instance=2)\n\ncontainer.register(Client)\n\nclient = container.resolve(Client)\nclient.get_number() # returns 2\n```\n\nWhen within a scope, Resolver returns the current scope rather than the container\n\n```python\nclass Client\n    def __init__(self, resolver: Resolver)\n        self.resolver = resolver\n\n    def get_number(self):\n        return self.resolver.resolve(int)\n\ncontainer.register(int, instance=2)\n\ncontainer.register(Client)\n\nclient = container.resolve(Client)\nclient.get_number() # returns 2\n\nwith container.get_scope() as scope:\n    scope.register(int, instance=10)\n    scoped_client = scope.resolve(Client)\n    scoped_client.get_number() # returns 10\n```\n\n\n## Modules (BETA feature)\n\n\nA module is a just a function that accepts a container, it can be used to set up common elements on the container\n\n```python\nclass ClientDependency\n    def get_int(self):\n        return 10\n\nclass Client\n    def __init__(self, dep: ClientDependency)\n        self.dep = dep\n\n    def get_number(self):\n        return self.dep.get_int()\n\ndef client_module(c: Container):\n    container = Container()\n    container.register(ClientDependency)\n    container.register(Client)\n\ncontainer.apply_module(client_module)\n\nclient = container.resolve(Client)\n\nclient.get_number() # returns 10\n```\n',
    'author': 'Peter Daly',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/peter-daly/clean_ioc',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
